﻿/*
 * TismTool © Tis Veugen, 2010
 * Application
 * 
 * TsmState.cs
 * version 4
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace TsmCode
{
  public enum TsmHandlerAction
  {
    NONE        = 0,
    NOK         = 1,
    FINAL       = 2,
    GUARD       = 4,
    OK          = 8,
    TRANSITION  = 16,
    ENTER       = 32,
    INITIAL     = 64,
    ENTRYPOINT  = 128,
    QUIT        = 256,
    EXITPOINT   = 512,
    TACTRANSITION = 1024
  }

  public class TsmHandler
  {
    public TsmHandlerAction Action = TsmHandlerAction.NONE;
    public int              Index  = -1;
    public TsmProspect      Prospect = TsmProspect.NOK;
    public bool             bTaken = false;
    public TsmClass         Owner;
    public TsmTrigger       Trigger;
    public TsmSubFsm        TacSubFsm;

    public TsmHandler() {}
    public TsmHandler( TsmClass oOwner )
    {
      Owner = oOwner;
    }

    public TsmHandler( TsmClass oOwner, TsmTrigger oTrigger )
    {
      Owner = oOwner;
      Trigger = oTrigger;
    }
  }

  public enum TsmProspect
  {
    NOK     = 0,
    OK      = 1,
    GUARD   = 4,
    ANY     = 16
  }

  public class TsmTransition
  {
    public bool     bVisit = false;
    public int      Unr = -1;
    public string   FromTo = "";

    public TsmTransition( int iUnr, string sFromTo )
    {
      Unr = iUnr;
      FromTo = sFromTo;
    }
  }

  public class TsmState : TsmBase
  {
    public bool             bVisit = false;
    public int              mNrTransitions = 0;
    public TsmTransition [] mTransitions = null;

    public virtual void AtEntry( TsmHandler oHandler )
    {
    }

    public virtual void AtExit( TsmHandler oHandler )
    {
    }

    public virtual void Initial( TsmHandler oHandler )
    {
    }

    public virtual void Entrypoint( TsmHandler oHandler )
    {
    }

    public virtual void ExecuteTrigger( TsmHandler oHandler )
    {
      // N.B. Some states may have no triggers
      oHandler.Action = TsmHandlerAction.NOK;
    }

    public virtual void ExecuteTriggerless( TsmHandler oHandler )
    {
      oHandler.Action = TsmHandlerAction.NOK;
    }

    public virtual bool ExecuteTransition( TsmHandler oHandler )
    {
      return true;
    }

    public void ExecuteHandler( TsmHandler oHandler )
    {
      bool bReady = false;

      do
      {
        switch ( oHandler.Action )
        {
        case TsmHandlerAction.TRANSITION:
        case TsmHandlerAction.TACTRANSITION:
          bReady = ExecuteTransition( oHandler );
          break;
        case TsmHandlerAction.EXITPOINT:
          oHandler.Action = TsmHandlerAction.TRANSITION;
          bReady = true;
          break;
        case TsmHandlerAction.ENTER:
          AtEntry( oHandler );
          Initial( oHandler );
          break;
        case TsmHandlerAction.INITIAL:
          Initial( oHandler );
          break;
        case TsmHandlerAction.ENTRYPOINT:
          AtEntry( oHandler );
          Entrypoint( oHandler );
          break;
        case TsmHandlerAction.OK:
        case TsmHandlerAction.GUARD:
        case TsmHandlerAction.NOK:  // Trigger to be handled at higher level
          bReady = true;
          break;
        case TsmHandlerAction.FINAL:
        case TsmHandlerAction.QUIT:
          ExecuteTriggerless( oHandler );
          bReady = true;
          break;
        default :
          Tsm.ReportBox(
            string.Format( "TsmHandlerAction '{0}' @ '{1}", oHandler.Action, this ) );
          bReady = true;
          break;
        }
      } while ( ! bReady );
    }

    public virtual TsmProspect GetProspect( int PortId, int Command )
    {
      return TsmProspect.NOK;
    }

    public virtual string TsmSnap()
    {
      return this.TsmName;
    }

    public virtual void TsmCheckVisit()
    {
    }
  }

  public class TsmPseudo : TsmState
  {
  }

  public class TsmFinal : TsmPseudo
  {
    public override void Initial( TsmHandler oHandler )
    {
      oHandler.Action = TsmHandlerAction.FINAL;
    }

    public override void ExecuteTrigger( TsmHandler oHandler )
    {
      oHandler.Action = TsmHandlerAction.FINAL;
    }

    public override void ExecuteTriggerless( TsmHandler oHandler )
    {
      oHandler.Action = TsmHandlerAction.FINAL;
    }
  }

  public class TsmRegion : TsmState
  {
    protected int         mNrStates = 0;
    protected TsmState [] mStates = null;
    protected TsmState    mCurrentState = null;
    protected bool        mbFinal = false;

    /**
     * This function should be overriden, or never be called.
     */
    public override void Initial( TsmHandler oHandler )
    {
      throw new Exception(
        string.Format( "9050: Region without Initial state @ {0}", this ) );
    }

    public override void ExecuteTrigger( TsmHandler oHandler )
    {
      if ( mbFinal )
      {
        oHandler.Action = TsmHandlerAction.FINAL;
        return;
      }

      TsmProspect Prospect = GetProspect( oHandler.Trigger.PortId, oHandler.Trigger.Command );
      switch ( Prospect )
      {
      case TsmProspect.NOK :
        oHandler.Action = TsmHandlerAction.NOK;
        return;
      case TsmProspect.ANY :
        if ( ( oHandler.Prospect & TsmProspect.OK ) != 0 )
        { /*
           * Execute the trigger at the higher level.
           */
          oHandler.Action = TsmHandlerAction.NOK;
          return;
        }
        break;
      default :
        break;
      }

      mCurrentState.ExecuteTrigger( oHandler );
      mCurrentState.ExecuteHandler( oHandler );
      ExecuteHandler( oHandler );
      mbFinal = ( oHandler.Action == TsmHandlerAction.FINAL );
    }

    public override void ExecuteTriggerless( TsmHandler oHandler )
    {
      // Prevent redefining oHandler.Action
    }

    public override void AtExit( TsmHandler oHandler )
    {
      if ( null != mCurrentState )
      {
        mCurrentState.AtExit( oHandler );
        Reset();
      }
    }

    public void Reset()
    {
      mCurrentState = null;
      mbFinal = false;
    }

    public override string TsmSnap()
    {
      if ( null == mCurrentState )
      {
        return "";
      }
      return mCurrentState.TsmSnap();
    }

    public override void TsmCheckVisit()
    {
      for ( int i=0; i<mNrStates; i++ )
      {
        mStates[i].TsmCheckVisit();
      }
    }
  }

  public class TsmCompound : TsmState
  {
    protected List<TsmRegion>   mRegions = new List<TsmRegion>();

    public override void Initial( TsmHandler oHandler )
    {
      ExecuteRegions( oHandler, false );
    }

    //todo: also others
    public override void ExecuteTrigger( TsmHandler oHandler )
    {
      TsmProspect ProspectSelf = GetProspect( oHandler.Trigger.PortId, oHandler.Trigger.Command );
      oHandler.Prospect |= ProspectSelf;

      ExecuteTriggerRegions( oHandler );
      if ( ! oHandler.bTaken )
      {
        // Process the trigger itself, if transition available.
        ExecuteTriggerCompound( oHandler );
      }
    }

    private void ExecuteTriggerRegions( TsmHandler oHandler )
    {
      ExecuteRegions( oHandler, true );
    }

    private void ExecuteRegions( TsmHandler oHandler, bool bTrigger )
    {
      TsmHandler hAll = new TsmHandler();
      foreach ( TsmRegion tRegion in mRegions )
      {
        TsmHandler h;
        
        if ( bTrigger )
        {
          h = new TsmHandler( oHandler.Owner, oHandler.Trigger );
          h.Prospect = oHandler.Prospect;
          tRegion.ExecuteTrigger( h );
        }
        else
        {
          h = new TsmHandler( oHandler.Owner );
          h.Action = TsmHandlerAction.ENTER;
          tRegion.ExecuteHandler( h );
        }

        switch ( h.Action )
        {
          case TsmHandlerAction.TRANSITION:
            if ( ( hAll.Action & (TsmHandlerAction.TRANSITION|TsmHandlerAction.TACTRANSITION) ) != 0 )
            {
              throw new Exception(
                string.Format( "9010: Multiple region transitions @ {0}", this ) );
            }
            hAll.Index = h.Index;
            break;
          case TsmHandlerAction.TACTRANSITION:
            if ( ( hAll.Action & (TsmHandlerAction.TRANSITION|TsmHandlerAction.TACTRANSITION) ) != 0 )
            {
              throw new Exception(
                string.Format( "9011: Multiple region transitions @ {0}", this ) );
            }
            hAll.Index = h.Index;
            hAll.TacSubFsm = h.TacSubFsm;
            break;
          default:
            break;
        }
        hAll.Action |= h.Action;
        hAll.bTaken |= h.bTaken;
      }

      HandlerTakeOver( oHandler, hAll );
      oHandler.bTaken |= hAll.bTaken;
    }

    private void HandlerTakeOver( TsmHandler oHandler, TsmHandler hAll )
    {
      if ( ( hAll.Action & TsmHandlerAction.TACTRANSITION ) != 0 )
      {
        oHandler.Action    = TsmHandlerAction.TACTRANSITION;
        oHandler.Index     = hAll.Index;
        oHandler.TacSubFsm = hAll.TacSubFsm;
      }
      else
      if ( ( hAll.Action & TsmHandlerAction.TRANSITION ) != 0 )
      {
        oHandler.Action = TsmHandlerAction.TRANSITION;
        oHandler.Index  = hAll.Index;
      }
      else
      if ( ( hAll.Action & TsmHandlerAction.OK ) != 0 )
      {
        oHandler.Action = TsmHandlerAction.OK;
      }
      else
      if ( ( hAll.Action & TsmHandlerAction.GUARD ) != 0 )
      {
        oHandler.Action = TsmHandlerAction.GUARD;
      }
      else
      if ( hAll.Action == TsmHandlerAction.FINAL )
      {
        oHandler.Action = TsmHandlerAction.FINAL;
      }
      else
      if ( hAll.Action == TsmHandlerAction.NOK )
      {
        oHandler.Action = TsmHandlerAction.NOK;
      }
      else
      if ( hAll.Action == (TsmHandlerAction.FINAL|TsmHandlerAction.NOK ))
      {
        oHandler.Action = TsmHandlerAction.NOK;
      }
      else
      {
        throw new Exception(
          string.Format( "9099: To be fixed @ {0}", this ) );
      }
    }

    public virtual void ExecuteTriggerCompound( TsmHandler oHandler )
    {
      // Override for all the compound triggers
      oHandler.Action = TsmHandlerAction.NOK;
    }

    public virtual void AtMyExit( TsmHandler oHandler )
    {
    }

    public void ExitRegions( TsmHandler oHandler )
    {
      foreach ( TsmRegion tRegion in mRegions )
      {
        tRegion.AtExit( oHandler );
      }
    }

    public override void AtExit( TsmHandler oHandler )
    {
      ExitRegions( oHandler );
      AtMyExit( oHandler );
    }

    public override string TsmSnap()
    {
      StringBuilder sb = new StringBuilder();
      sb.Append( "[" + this.TsmName + ":" );
      foreach ( TsmRegion tRegion in mRegions )
      {
        sb.Append( "|" + tRegion.TsmSnap() );
      }
      return sb.ToString() + "]";
    }

    public override void TsmCheckVisit()
    {
      foreach ( TsmRegion tRegion in mRegions )
      {
        tRegion.TsmCheckVisit();
      }
    }
  }

  public class TsmFsm : TsmCompound
  {
    public void ExecuteTriggerFsm( TsmHandler oHandler )
    {
      ExecuteTrigger( oHandler );
      if ( ! oHandler.bTaken )
      {
        TsmSupervisor.TsmRecorder.UnhandledFunction( oHandler.Trigger.ToString() + " @ " +
          this.TsmName + " (" + this.ToString() + ")" );
      }
    }

    public override void ExecuteTriggerless( TsmHandler oHandler )
    {
      oHandler.Action = TsmHandlerAction.QUIT;
    }

    public void TsmReset( TsmHandler oHandler )
    {
      AtExit( oHandler );
      Initial( oHandler );
    }

    public override string TsmSnap()
    {
      return "Fsm:" + base.TsmSnap();
    }
  }

  public class TsmLeaf : TsmState
  {
    public override void AtEntry( TsmHandler oHandler )
    {
      // No need to call.
    }
    
    public virtual void AtDo()
    {
      // No need to call.
    }
    
    public override void Initial( TsmHandler oHandler )
    {
      oHandler.Action = TsmHandlerAction.OK;
    }

    /*
     * Results of oHandler.Action :
     * NOK: no triggers, or triggerless
     * GUARD: trigger, guarded
     * OK: internal transition (even ignored transition)
     * TRANSITION: trigger honoured
     * 
     * Completion event after:
     * 1. firstly entering this state, after AtEntry() and AtDo()
     * 2. internal transition, and AtDo()
     */
    public override void ExecuteTrigger( TsmHandler oHandler )
    {
      oHandler.Action = TsmHandlerAction.NOK;
    }
  }

  public class TsmSubFsm : TsmState
  {
    protected TsmFsm oFsm;

    public override void ExecuteTrigger( TsmHandler oHandler )
    {
      TsmProspect ProspectSelf = GetProspect( oHandler.Trigger.PortId, oHandler.Trigger.Command );
      oHandler.Prospect |= ProspectSelf;

      oFsm.ExecuteTrigger( oHandler );
      oFsm.ExecuteHandler( oHandler );

      if ( oHandler.Action == TsmHandlerAction.FINAL )
      {
        ExecuteTriggerless( oHandler );
      }
      else
      if ( oHandler.Action == TsmHandlerAction.QUIT )
      {
        ExecuteTriggerless( oHandler );
      }
      else
      if ( ! oHandler.bTaken )
      {
        // Process the trigger itself, if transition available.
        ExecuteTriggerSubFsm( oHandler );
      }
    }

    public virtual void ExecuteTriggerSubFsm( TsmHandler oHandler )
    {
      // Override for all the SubFsm triggers
      oHandler.Action = TsmHandlerAction.NOK;
    }

    public override void Initial( TsmHandler oHandler )
    {
      oHandler.Action = TsmHandlerAction.ENTER;
      oFsm.ExecuteHandler( oHandler );

      if ( oHandler.Action == TsmHandlerAction.FINAL )
      {
        ExecuteTriggerless( oHandler );
      }

      if ( oHandler.Action == TsmHandlerAction.QUIT )
      {
        ExecuteTriggerless( oHandler );
      }
    }

    public virtual void AtMyExit( TsmHandler oHandler )
    {
    }

    public override void AtExit( TsmHandler oHandler )
    {
      oFsm.AtExit( oHandler );
      this.AtMyExit( oHandler );
    }

    public TsmFsm SwapFsms( TsmFsm Target2Source )
    {
      TsmFsm h = oFsm;
      TsmBase hp = oFsm.TsmParent;
      oFsm = Target2Source;
      h.TsmParent = Target2Source.TsmParent;
      oFsm.TsmParent = hp;
      return h;
    }

    public override string TsmSnap()
    {
      return this.TsmName + "+" + oFsm.TsmSnap();
    }

    public override void TsmCheckVisit()
    {
      oFsm.TsmCheckVisit();
    }
  }
}
