﻿/*
 * TismTool © Tis Veugen, 2010
 * Application
 *
 * TsmSupervisor.cs
 * version 6
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Xml;
using System.Xml.Linq;
using System.Text;
using System.Threading;
using System.Runtime.Remoting.Contexts;
using System.Diagnostics; // Process
using TsmCode;

namespace TsmSupervisor
{

  public class TsmRzVouzSem
  {
    private Semaphore mSem;
    private bool      mbRendezVous = false;

    public TsmRzVouzSem()
    {
      mSem = new Semaphore( 0, 1 );
    }

    public void WaitOne()
    {
      //Tsm.Assert( ! mbRendezVous );
      mbRendezVous = true;
      mSem.WaitOne();
      mbRendezVous = false;
    }

    public void Release()
    {
      //Tsm.Assert( mbRendezVous );
      mSem.Release();
    }

    public void Reset()
    {
      if ( mbRendezVous )
      {
        mbRendezVous = false;
        // Easiest way to reset the semaphore.
        mSem = new Semaphore( 0, 1 );
      }
    }
  }

  public class TsmTriggerSem
  {
    Semaphore mSem;

    public TsmTriggerSem()
    {
      mSem = new Semaphore( 0, 1 );
    }

    public void WaitOne()
    {
      mSem.WaitOne();
    }

    public void Release()
    {
      try
      {
        mSem.Release();
      }
      catch ( SemaphoreFullException )
      { /*
         * Allowed to prevent excessive trigger looping.
         */
      }
    }
  }

  [Synchronization]
  public class TsmExtQueue : Queue<TsmEnvelope>
  {
  }

  [Synchronization]
  public class TsmIntQueue : Queue<TsmEnvelope>
  {
    private TsmThread mTsmThread;

    public TsmIntQueue( TsmThread oTsmThread )
    {
      mTsmThread = oTsmThread;
    }

    public void Put( TsmEnvelope oEnvelope )
    {
      this.Enqueue( oEnvelope );
    }
  }

  public enum TsmThreadState
  {
    RUNNING,
    WAITING_MSG,
    WAITING_RV,
    WAITING_RV2,
    INTERRUPTED
  }

  public class TsmThread
  {
    public TsmThreadType    mType;
    public int              mSequence;
    protected Thread        mThread;
    protected TsmExtQueue   mExternalQueue = new TsmExtQueue();
    protected TsmIntQueue   mInternalQueue;
    protected Mutex         mReentrantMutex = new Mutex();
    protected bool          mbReentrant = false;
    protected TsmRzVouzSem  mRendezVousSem = new TsmRzVouzSem();
    protected TsmTriggerSem mTriggerSem = new TsmTriggerSem();

    public TsmThread() : this( TsmThreadType.THREAD )
    {
    }

    public TsmThread( TsmThreadType Type )
    {
      mType = Type;
      mInternalQueue = new TsmIntQueue( this );
    }

    public TsmIntQueue IntQueue
    {
      get { return mInternalQueue; }
    }

    public Thread Thread
    {
      get { return mThread; }
    }

    public bool HasMessage
    {
      get { return ( mExternalQueue.Count + mInternalQueue.Count > 0 ); }
    }

    public virtual void Prepare()
    {
      mThread = new Thread( TsmThread.Worker );
      mSequence = TsmSupervisor.Register( this, mType );
      mThread.Name = "Tsm_" + mSequence;
    }

    public virtual void Run()
    {
      mThread.Start( this );
    }

    public virtual void Halt()
    {
      mThread.Abort();
      mThread.Join();
    }
    
    public virtual void Interrupt()
    {
      mThread.Interrupt();
    }

    public virtual void Cease()
    {
    }

    public void ReentrantLock()
    {
      //Tsm.Assert( ! mbReentrant );
      mReentrantMutex.WaitOne();
      mbReentrant = true;
    }

    public void ReentrantUnlock()
    {
      mbReentrant = false;
      mReentrantMutex.ReleaseMutex();
    }

    public void RendezVousWait()
    {
      mRendezVousSem.WaitOne();
    }

    public void RendezVousNotify()
    {
      mRendezVousSem.Release();
    }

    public void TriggerWait()
    {
      mTriggerSem.WaitOne();
    }

    public void TriggerNotify()
    {
      mTriggerSem.Release();
    }

    public void ASyncCall( TsmEnvelope oEnvelope )
    {
      mExternalQueue.Enqueue( oEnvelope );

      TriggerNotify();
    }

    public void SyncCall( TsmEnvelope oEnvelope )
    {
      ReentrantLock();

      oEnvelope.Trigger.RvState = TsmRvState.WAIT;
      mExternalQueue.Enqueue( oEnvelope );

      TriggerNotify();
      RendezVousWait();

      ReentrantUnlock();
    }

    protected TsmRvState Handle( TsmEnvelope oEnvelope )
    {
      oEnvelope.Target.ExecuteTrigger( oEnvelope.Trigger );
      if ( oEnvelope.Trigger.RvState == TsmRvState.READY )
      {
        RendezVousNotify();
        oEnvelope.Trigger.RvState = TsmRvState.IDLE;
      }
      return oEnvelope.Trigger.RvState;
    }

    public static void Worker( object oObject )
    {
      TsmThread   oTsmThread = (TsmThread) oObject;
      TsmEnvelope lEnvelope;
      TsmRvState  lRvState = TsmRvState.IDLE;
      bool        bMsg = false;

      for ( ; ; )
      {
        oTsmThread.TriggerWait();

        do
        {
          bMsg = false;
          try
          {
            lEnvelope = oTsmThread.mExternalQueue.Dequeue();
            if ( lEnvelope.Trigger.RvState == TsmRvState.WAIT )
            {
              lRvState = TsmRvState.WAIT;
            }
            else
            {
              lEnvelope.Trigger.RvState = lRvState;
            }
            lRvState = oTsmThread.Handle( lEnvelope );
            bMsg = true;
          }
          catch ( InvalidOperationException )
          {
            // Console.WriteLine( "No message" );
          }
          catch ( Exception e )
          {
            Console.WriteLine( e.Message );
            Tsm.Lgl( 1, e.Message );
            Tsm.Lgl( 1, "@Q : " + e.StackTrace );
          }

          for ( ; ; )
          {
            try
            {
              lEnvelope = oTsmThread.mInternalQueue.Dequeue();
              lEnvelope.Trigger.RvState = lRvState;
              lRvState = oTsmThread.Handle( lEnvelope );
              bMsg = true;
            }
            catch ( InvalidOperationException )
            {
              // Console.WriteLine( "No more messages" );
              break;
            }
            catch ( Exception e )
            {
              Console.WriteLine( e.Message );
              Tsm.Lgl( 1, e.Message );
              Tsm.Lgl( 1, "@S : " + e.StackTrace );
              break;
            }
          }
        } while ( bMsg );
      }
    }
  }

  public class TsmCbThread : TsmThread
  {
    public TsmCbThread() : base( TsmThreadType.CB_THREAD )
    {
    }

    public override void Prepare()
    {
      mThread = new Thread( TsmCbThread.Worker );
      mSequence = TsmSupervisor.Register( this, mType );
      mThread.Name = "Tsm_Cb";
    }

    public void NativeCall( TsmEnvelope oEnvelope )
    {
      mExternalQueue.Enqueue( oEnvelope );
    }

    protected void CbHandle( TsmEnvelope oEnvelope )
    {
      oEnvelope.Target.ExecuteNative( oEnvelope.Trigger.Command );
    }

    public new static void Worker( object oObject )
    {
      TsmCbThread oTsmThread = (TsmCbThread) oObject;
      TsmEnvelope lEnvelope;
      bool        bMsg = false;

      for ( ; ; )
      {
        oTsmThread.TriggerWait();

        bMsg = false;
        try
        {
          lEnvelope = oTsmThread.mExternalQueue.Dequeue();
          oTsmThread.CbHandle( lEnvelope );
          bMsg = true;
        }
        catch ( InvalidOperationException )
        {
          // Console.WriteLine( "No message" );
        }
        catch ( Exception e )
        {
          Console.WriteLine( e.Message );
          Tsm.Lgl( 1, e.Message );
          Tsm.Lgl( 1, "@q : " + e.StackTrace );
        }

        if ( ! bMsg )
        {
          try
          {
            lEnvelope = oTsmThread.mInternalQueue.Dequeue();
            oTsmThread.Handle( lEnvelope );
          }
          catch ( InvalidOperationException )
          {
            // Console.WriteLine( "No more messages" );
          }
          catch ( Exception e )
          {
            Console.WriteLine( e.Message );
            Tsm.Lgl( 1, e.Message );
            Tsm.Lgl( 1, "@s : " + e.StackTrace );
          }
        }
      }
    }
  }

  public class TsmActivity
  {
    private int         mNrAnnounce = 0;
    private int         mNrAbandon = 0;
    private Semaphore   mSem = new Semaphore( 0, 16 );

    public void Wait()
    {
      while ( mNrAnnounce != mNrAbandon )
      {
        mSem.WaitOne();
      }
    }

    public void Announce()
    {
      mNrAnnounce++;
    }

    public void Abandon()
    {
      mNrAbandon++;
      mSem.Release();
    }
  }

  public enum TsmThreadType
  {
    SUPERVISOR,
    CB_THREAD,
    THREAD
  }

  [Synchronization]
  public partial class TsmSupervisor
  {
    private static TsmActivity   mActivity;
    private static int           mThreadId = 0;
    private static TsmCbThread   mCbThread;

    public static void Initialise()
    {
      mActivity = new TsmActivity();
      Register( null, TsmThreadType.SUPERVISOR );

      mCbThread = new TsmCbThread();
      mCbThread.Prepare();
      mCbThread.Run();
    }

    public static void Terminate()
    {
      mCbThread.Halt();
      mCbThread.Cease();
      Tsm.Close();
    }

    public static TsmIntQueue GetIntQueue()
    {
      return mCbThread.IntQueue;
    }

    public static void Announce()
    {
      mActivity.Announce();
    }

    public static void Abandon()
    {
      mActivity.Abandon();
    }

    public static int Register( TsmThread oTsmThread )
    {
      return Register( oTsmThread, TsmThreadType.THREAD );
    }

    public static int Register( TsmThread oTsmThread, TsmThreadType eType )
    {
      return mThreadId++;
    }

    public static void Execute()
    {
      mActivity.Wait();
    }
  }

  [Synchronization]
  public class TsmRecorder
  {
    public static int           NumberOfUnhandledFunctions = 0;

    public static void UnhandledFunction( string sEvent )
    {
      NumberOfUnhandledFunctions++;
      Show( "*** UNHANDLED FUNCTION *** : " + sEvent );
    }

    public static void Show( string sEvent )
    {
      Tsm.Log( sEvent );
    }
  }
}
